@@ -27,6 +27,9 @@
 
 #include <icon-helpers.h>
 
+#define GNOME_DESKTOP_USE_UNSTABLE_API 1
+#include <libgnome-desktop/gnome-desktop-thumbnail.h>
+
 #define THUMB_SEARCH_SIZE     256
 #define THUMB_SEARCH_HEIGHT   (THUMB_SEARCH_SIZE / 4 * 3)
 
@@ -38,21 +41,19 @@ typedef enum {
 	NUM_ICONS
 } IconType;
 
+static GnomeDesktopThumbnailFactory *factory;
+static GThreadPool *thumbnail_pool;
 static GdkPixbuf *icons[NUM_ICONS];
 static GHashTable *cache_thumbnails; /* key=url, value=GdkPixbuf */
 
 GdkPixbuf *
-totem_grilo_get_thumbnail_finish (GAsyncResult  *res,
+totem_grilo_get_thumbnail_finish (GObject       *source_object,
+				  GAsyncResult  *res,
 				  GError       **error)
 {
-	GSimpleAsyncResult *simple = G_SIMPLE_ASYNC_RESULT (res);
+	g_return_val_if_fail (g_task_is_valid (res, source_object), NULL);
 
-	g_warn_if_fail (g_simple_async_result_get_source_tag (simple) == totem_grilo_get_thumbnail);
-
-	if (g_simple_async_result_propagate_error (simple, error))
-		return NULL;
-
-	return g_simple_async_result_get_op_res_gpointer (simple);
+	return g_task_propagate_pointer (G_TASK (res), error);
 }
 
 static void
@@ -60,30 +61,28 @@ load_thumbnail_cb (GObject *source_object,
 		   GAsyncResult *res,
 		   gpointer user_data)
 {
-	GSimpleAsyncResult *simple = G_SIMPLE_ASYNC_RESULT (user_data);
+	GTask *task = user_data;
 	GdkPixbuf *pixbuf;
 	GError *error = NULL;
 	const GFile *file;
 
 	pixbuf = gdk_pixbuf_new_from_stream_finish (res, &error);
 	if (!pixbuf) {
-		g_simple_async_result_take_error (simple, error);
-		g_simple_async_result_complete_in_idle (simple);
-		g_object_unref (res);
+		g_task_return_error (task, error);
+		g_object_unref (task);
 		return;
 	}
 
 	/* Cache it */
-	file = g_object_get_data (source_object, "file");
+	file = g_task_get_task_data (task);
 	if (file) {
 		g_hash_table_insert (cache_thumbnails,
 				     g_file_get_uri (G_FILE (file)),
 				     g_object_ref (pixbuf));
 	}
 
-	g_simple_async_result_set_op_res_gpointer (simple, pixbuf, NULL);
-	g_simple_async_result_complete_in_idle (simple);
-	g_object_unref (simple);
+	g_task_return_pointer (task, pixbuf, g_object_unref);
+	g_object_unref (task);
 }
 
 static void
@@ -91,43 +90,130 @@ get_stream_thumbnail_cb (GObject *source_object,
 			 GAsyncResult *res,
 			 gpointer user_data)
 {
-	GSimpleAsyncResult *simple = G_SIMPLE_ASYNC_RESULT (user_data);
+	GTask *task = user_data;
 	GFileInputStream *stream;
-	GCancellable *cancellable;
 	GError *error = NULL;
 
 	stream = g_file_read_finish (G_FILE (source_object), res, &error);
 	if (!stream) {
-		g_simple_async_result_take_error (simple, error);
-		g_simple_async_result_complete_in_idle (simple);
+		g_task_return_error (task, error);
+		g_object_unref (task);
 		return;
 	}
 
-	cancellable = g_object_get_data (G_OBJECT (simple), "cancellable");
 	gdk_pixbuf_new_from_stream_at_scale_async (G_INPUT_STREAM (stream),
 						   THUMB_SEARCH_SIZE,
 						   THUMB_SEARCH_HEIGHT,
 						   TRUE,
-						   cancellable,
+						   g_task_get_cancellable (task),
 						   load_thumbnail_cb,
-						   simple);
+						   task);
 	g_object_unref (G_OBJECT (stream));
 }
 
+static void
+thumbnail_media_async_thread (GTask    *task,
+			      gpointer  user_data)
+{
+	GrlMedia *media;
+	GdkPixbuf *pixbuf;
+	const char *uri;
+	GDateTime *mtime;
+
+	if (g_task_return_error_if_cancelled (task)) {
+		g_object_unref (task);
+		return;
+	}
+
+	media = GRL_MEDIA (g_task_get_source_object (task));
+	uri = grl_media_get_url (media);
+	mtime = grl_media_get_modification_date (media);
+
+	if (!uri || !mtime) {
+		g_task_return_new_error (task, G_IO_ERROR, G_IO_ERROR_FAILED, "URI or mtime missing");
+		g_object_unref (task);
+		return;
+	}
+
+	pixbuf = gnome_desktop_thumbnail_factory_generate_thumbnail (factory, uri, "video/x-totem-stream");
+
+	if (!pixbuf) {
+		g_task_return_new_error (task, G_IO_ERROR, G_IO_ERROR_FAILED, "Thumbnailing failed");
+		g_object_unref (task);
+		return;
+	}
+
+	gnome_desktop_thumbnail_factory_save_thumbnail (factory, pixbuf, uri, g_date_time_to_unix (mtime));
+	g_task_return_pointer (task, pixbuf, g_object_unref);
+	g_object_unref (task);
+}
+
+static void
+totem_grilo_thumbnail_media (GrlMedia            *media,
+			     GCancellable        *cancellable,
+			     GAsyncReadyCallback  callback,
+			     gpointer             user_data)
+{
+	GTask *task;
+
+	task = g_task_new (media, cancellable, callback, user_data);
+	g_task_set_priority (task, G_PRIORITY_LOW);
+	g_thread_pool_push (thumbnail_pool, task, NULL);
+}
+
+static GdkPixbuf *
+totem_grilo_thumbnail_media_finish (GrlMedia      *media,
+				    GAsyncResult  *res,
+				    GError       **error)
+{
+	g_return_val_if_fail (g_task_is_valid (res, media), NULL);
+
+	return g_task_propagate_pointer (G_TASK (res), error);
+}
+
+static void
+thumbnail_media_cb (GObject      *source_object,
+		    GAsyncResult *res,
+		    gpointer      user_data)
+{
+	GTask *task = user_data;
+	GdkPixbuf *pixbuf;
+	GError *error = NULL;
+
+	pixbuf = totem_grilo_thumbnail_media_finish (GRL_MEDIA (source_object), res, &error);
+	if (!pixbuf)
+		g_task_return_error (task, error);
+	else
+		g_task_return_pointer (task, pixbuf, g_object_unref);
+	g_object_unref (task);
+}
+
 void
 totem_grilo_get_thumbnail (GObject             *object,
 			   GCancellable        *cancellable,
 			   GAsyncReadyCallback  callback,
 			   gpointer             user_data)
 {
-	GSimpleAsyncResult *simple;
+	GTask *task;
 	const char *url_thumb = NULL;
-	const GdkPixbuf *thumbnail;
+	const GdkPixbuf *thumbnail = NULL;
 	GFile *file;
 
-	if (GRL_IS_MEDIA (object))
+	task = g_task_new (G_OBJECT (object),
+			   cancellable,
+			   callback,
+			   user_data);
+
+	if (GRL_IS_MEDIA (object)) {
 		url_thumb = grl_media_get_thumbnail (GRL_MEDIA (object));
-	else if (GRL_IS_SOURCE (object)) {
+		if (!url_thumb && g_strcmp0 (grl_media_get_source (GRL_MEDIA (object)), "grl-tracker-source") == 0) {
+			totem_grilo_thumbnail_media (GRL_MEDIA (object),
+						     cancellable,
+						     thumbnail_media_cb,
+						     task);
+			return;
+		}
+	} else if (GRL_IS_SOURCE (object)) {
 		GIcon *icon;
 
 		icon = grl_source_get_icon (GRL_SOURCE (object));
@@ -137,32 +223,28 @@ totem_grilo_get_thumbnail (GObject             *object,
 			file = g_file_icon_get_file (G_FILE_ICON (icon));
 			url_thumb = g_file_get_uri (file);
 			g_object_unref (file);
-		} else {
-			//FIXME
-			return;
 		}
 	}
-	g_return_if_fail (url_thumb != NULL);
-
-	simple = g_simple_async_result_new (G_OBJECT (object),
-					    callback,
-					    user_data,
-					    totem_grilo_get_thumbnail);
+	if (url_thumb == NULL) {
+		g_task_return_pointer (task, NULL, NULL);
+		g_object_unref (task);
+		return;
+	}
 
 	/* Check cache */
 	thumbnail = g_hash_table_lookup (cache_thumbnails, url_thumb);
 	if (thumbnail) {
-		g_simple_async_result_set_op_res_gpointer (simple, g_object_ref (G_OBJECT (thumbnail)), NULL);
-		g_simple_async_result_complete_in_idle (simple);
-		g_object_unref (simple);
+		g_task_return_pointer (task,
+				       g_object_ref (G_OBJECT (thumbnail)),
+				       g_object_unref);
+		g_object_unref (task);
 		return;
 	}
 
 	file = g_file_new_for_uri (url_thumb);
-	g_object_set_data_full (G_OBJECT (simple), "file", file, g_object_unref);
-	g_object_set_data (G_OBJECT (simple), "cancellable", cancellable);
+	g_task_set_task_data (task, file, g_object_unref);
 	g_file_read_async (file, G_PRIORITY_DEFAULT, NULL,
-			   get_stream_thumbnail_cb, simple);
+			   get_stream_thumbnail_cb, task);
 }
 
 static void
@@ -237,7 +319,8 @@ totem_grilo_get_icon (GrlMedia *media,
 	if (GRL_IS_MEDIA_BOX (media)) {
 		return g_object_ref (icons[ICON_BOX]);
 	} else {
-		if (grl_media_get_thumbnail (media)) {
+		if (grl_media_get_thumbnail (media) ||
+		    g_strcmp0 (grl_media_get_source (media), "grl-tracker-source") == 0) {
 			*thumbnailing = TRUE;
 			return g_object_ref (icons[ICON_VIDEO_THUMBNAILING]);
 		} else {
@@ -276,6 +359,9 @@ totem_grilo_clear_icons (void)
 		g_clear_object (&icons[i]);
 
 	g_clear_pointer (&cache_thumbnails, g_hash_table_destroy);
+	g_clear_object (&factory);
+	g_thread_pool_free (thumbnail_pool, TRUE, FALSE);
+	thumbnail_pool = NULL;
 }
 
 void
@@ -290,4 +376,7 @@ totem_grilo_setup_icons (Totem *totem)
 						  g_str_equal,
 						  g_free,
 						  g_object_unref);
+
+	factory = gnome_desktop_thumbnail_factory_new (GNOME_DESKTOP_THUMBNAIL_SIZE_LARGE);
+	thumbnail_pool = g_thread_pool_new ((GFunc) thumbnail_media_async_thread, NULL, 5, TRUE, NULL);
 }
